<?php

namespace Drupal\entity_sync\FieldTransformer;

use Drupal\entity_sync\Exception\InvalidConfigurationException;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Plugin\PluginBase as CorePluginBase;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\Validator\ValidatorInterface;
use Symfony\Component\Validator\ConstraintViolationListInterface;

/**
 * Default base class for field transformer plugin implementations.
 *
 * Supported configuration properties provided for all plugins:
 * Applicable to import configurations only:
 * - setter: (string, optional) The name of a setter method to use when setting
 *   the value of the property/field on the local entity. If not defined, the
 *   `set` method provided by the Field API will be used. If defined, the method
 *   must exist on the local entity.
 *
 * Note that default values (here or in inheriting classes) should be provided
 * in `defaultConfiguration` for properties that are optional - and only for
 * those. Setting default values for properties that are required nullifies the
 * validation and that can lead to mistakes and unintended behavior - there will
 * always be a value (the provided default) and validation will always pass even
 * when the configuration does not provide values for the properties that are
 * meant to be required.
 *
 * The `defaultConfiguration` method is meant to do just that - provide default
 * values for the properties. Which properties are supported should be indicated
 * in the class description, like above.
 */
abstract class PluginBase extends CorePluginBase implements PluginInterface {

  /**
   * The constraint plugin manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $constraintManager;

  /**
   * The constraint validator.
   *
   * @var \Symfony\Component\Validator\Validator\ValidatorInterface
   */
  protected $constraintValidator;

  /**
   * Constructs a new PluginBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Symfony\Component\Validator\Validator\ValidatorInterface $constraint_validator
   *   The constraint validator.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $constraint_manager
   *   The constraint plugin manager.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    ValidatorInterface $constraint_validator,
    PluginManagerInterface $constraint_manager
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->setConfiguration($configuration);
    $this->validateConfiguration();

    $this->constraintValidator = $constraint_validator;
    $this->constraintManager = $constraint_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    ValidatorInterface $constraint_validator,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $constraint_validator,
      $container->get('plugin.manager.entity_sync_validation_constraint')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    return [
      'module' => [$this->pluginDefinition['provider']],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'setter' => NULL,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration + $this->defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function import(
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    $this->setLocalEntityValue(
      $local_entity,
      $field_info,
      $this->processImportedValue(
        $remote_entity,
        $local_entity,
        $field_info,
        $context
      )
    );
  }

  /**
   * Validates the plugin configuration.
   *
   * This method is called in the constructor so that errors in the plugin's
   * configuration are caught.
   *
   * @throws \Drupal\entity_sync\Exception\InvalidConfigurationException
   *   When the plugin is improperly configured.
   *
   * @I Validate configuration upon import when supported by core
   *    type     : bug
   *    priority : low
   *    labels   : config, field-transformer, validation
   * @I Support validating configuration property types e.g. string, int etc.
   *    type     : bug
   *    priority : low
   *    labels   : config, field-transformer, validation
   */
  protected function validateConfiguration() {
    // Nothing to validate by default. Plugin implementations may define
    // additional configuration properties which they can validate by overriding
    // this method.
  }

  /**
   * Validates that the given properties are present in the configuration.
   *
   * @param array $keys
   *   The required configuration property keys.
   *
   * @throws \Drupal\entity_sync\Exception\InvalidConfigurationException
   *   When one or more required properties are missing.
   */
  protected function validateConfigurationRequiredProperties(array $keys) {
    $missing_keys = array_diff($keys, array_keys($this->configuration));
    if (!$missing_keys) {
      return;
    }

    throw new InvalidConfigurationException(sprintf(
      'The following configuration properties are required: %s',
      implode(', ', $missing_keys)
    ));
  }

  /**
   * Sets the local entity field to the given value.
   *
   * If a setter method is defined in the plugin configuration, it uses
   * that. Otherwise it uses the `set` method provided by the
   * `FieldItemListInterface` API.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param array $field_info
   *   The field info.
   *   See \Drupal\entity_sync\Import\Event\FieldMapping::fieldMapping.
   * @param mixed $value
   *   The value to set the field to.
   */
  protected function setLocalEntityValue(
    ContentEntityInterface $local_entity,
    array $field_info,
    $value
  ) {
    if (!isset($this->configuration['setter'])) {
      $local_entity->set(
        $field_info['machine_name'],
        $value
      );
      return;
    }

    // We cannot easily validate this in the `validateConfiguration` method
    // because we don't have the entity available there. We would need to have
    // the synchronization available, load the entity type, get the entity class
    // and find whether it implements the method. Even that would be the correct
    // place to do such validation, it's not worth the complexity.
    // Let's do it here instead.
    if (!method_exists($local_entity, $this->configuration['setter'])) {
      throw new InvalidConfigurationException(sprintf(
        'Unknown method "%s" on entity of type "%s".',
        $this->configuration['setter'],
        $local_entity->getEntityTypeId()
      ));
    }

    $local_entity->{$this->configuration['setter']}($value);
  }

  /**
   * Process the value being imported before setting it in the entity field.
   *
   * Processing includes:
   * - Pre-transformation validation.
   * - Transformation.
   * - Post-transformation validation.
   *
   * @param object $remote_entity
   *   The remote entity.
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param array $field_info
   *   The field info.
   *   See \Drupal\entity_sync\Import\Event\FieldMapping::fieldMapping.
   * @param array $context
   *   An associative array containing context related to the circumstances of
   *   the operation.
   *
   * @return mixed
   *   The transformed value.
   */
  protected function processImportedValue(
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    $original_value = $remote_entity->{$field_info['remote_name']};

    $this->validate($original_value, 'pre_transformation');

    $transformed_value = $this->transformImportedValue(
      $original_value,
      $remote_entity,
      $local_entity,
      $field_info,
      $context
    );

    $this->validate($transformed_value, 'post_transformation');

    return $transformed_value;
  }

  /**
   * Transforms the value being imported.
   *
   * The public API of the field transformer plugin requires the `import` method
   * to be implemented. This base class provides a default implementation of
   * that method that performs the following common actions so that plugin
   * implementations do not have to:
   * - Validates the original value before transformation.
   * - Transforms the value calling this method.
   * - Validates the transformed value.
   * - Sets the field to the transformed value.
   *
   * With this, most plugin implementations will only need to implement this
   * method (`transformImportedValue`), and possibly the `setLocalEntityValue`
   * method if you need to customize the way the field value is set on the
   * entity.
   *
   * Note that this method (`transformImportedValue`) is not declared `abstract`
   * because plugin implementations could have a reason to override directly the
   * `import` method; in that case they may not need to implement
   * `transformImportedValue`.
   *
   * If you do have a reason to override the `import` method, make sure that you
   * take into account the actions listed above so that they are not ommitted.
   *
   * @param mixed $value
   *   The original value.
   * @param object $remote_entity
   *   The remote entity.
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param array $field_info
   *   The field info.
   *   See \Drupal\entity_sync\Import\Event\FieldMapping::fieldMapping.
   * @param array $context
   *   An associative array containing context related to the circumstances of
   *   the operation.
   *
   * @return mixed
   *   The transformed value.
   */
  protected function transformImportedValue(
    $value,
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    throw new \Exception('An implementation of the `transformImportedValue` method must be provided.');
  }

  /**
   * Validates the given value against the constraints of the given group.
   *
   * The default behavior in the case that validation fails is to throw an
   * exception causing the import of the entire entity to fail. Support for
   * failure handler should be provided in the future.
   *
   * @param mixed $value
   *   The value to validate.
   * @param string $group
   *   The constraint group.
   *
   * @throws \RuntimeException
   *   When validation fails.
   *
   * @I Collect all errors from all constraints before failing
   *    type     : improvement
   *    priority : normal
   *    labels   : field-transformer, validation
   * @I Support validation failure handlers
   *    type     : improvement
   *    priority : normal
   *    labels   : field-transformer, validation
   *    notes    : Possible failure handlers are fail, skip and nullify, all
   *               with a configuration option to log a message.
   */
  protected function validate($value, string $group) {
    if (empty($this->configuration['constraints'])) {
      return;
    }

    $constraints = $this->getConstraintsForGroup($group);
    if (!$constraints) {
      return;
    }

    foreach ($constraints as $constraint) {
      $errors = $this->constraintValidator->validate($value, $constraint);
      if (!$errors->count()) {
        continue;
      }

      throw new \RuntimeException(
        $this->buildExceptionMessage($group, $errors)
      );
    }
  }

  /**
   * Returns the instantiated constraint plugins for the given group.
   *
   * @param string $group
   *   The constraint group.
   *
   * @return \Drupal\entity_sync\FieldTransformer\PluginInterface[]
   *   The constraints for the given group.
   */
  protected function getConstraintsForGroup(string $group) {
    $constraints = array_filter(
      $this->configuration['constraints'],
      function ($constraint) use ($group) {
        if (empty($constraint['groups'])) {
          return FALSE;
        }

        return in_array($group, $constraint['groups']);
      }
    );

    return array_map(
      function ($constraint) {
        return $this->constraintManager->createInstance(
          $constraint['id'],
          $constraint['configuration'] ?? []
        );
      },
      $constraints
    );
  }

  /**
   * Builds the exception message in the case we have validation errors.
   *
   * @param string $group
   *   The constraint group.
   * @param \Symfony\Component\Validator\ConstraintViolationListInterface $errors
   *   The constraint violation list.
   *
   * @return string
   *   The exception message.
   */
  protected function buildExceptionMessage(
    string $group,
    ConstraintViolationListInterface $errors
  ) {
    $messages = [];
    foreach ($errors as $index => $error) {
      $messages[] = (string) ++$index . ') ' . $error->getMessage();
    }

    return sprintf(
      'The following errors occurred while validating the value against the "%s" constraint group: "%s"',
      $group,
      implode(' ', $messages)
    );
  }

}
