<?php

namespace Drupal\entity_sync\Plugin\EntitySync\FieldTransformer;

use Drupal\entity_sync\Exception\InvalidConfigurationException;
use Drupal\entity_sync\FieldTransformer\PluginBase;
use Drupal\Core\Entity\ContentEntityInterface;

/**
 * Transformer that copies the value being imported/exported.
 *
 * Supported configuration properties:
 * - cast_to: (string, optional) The type to cast to the value being
 *   transformed. Supported types are the ones that are supported by PHP
 *   type casting:
 *   - int/integer
 *   - bool/boolean
 *   - float/double/real
 *   - string
 *   - array
 *   - object
 *   - unset
 *
 * @EntitySyncFieldTransformer(
 *   id = "copy_value"
 * )
 *
 * @link https://www.php.net/manual/en/language.types.type-juggling.php
 */
class CopyValue extends PluginBase {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'cast_to' => NULL,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  protected function validateConfiguration() {
    if (!isset($this->configuration['cast_to'])) {
      return;
    }

    $supported_types = [
      'int',
      'integer',
      'bool',
      'boolean',
      'float',
      'double',
      'real',
      'string',
      'array',
      'object',
      'unset',
    ];
    if (in_array($this->configuration['cast_to'], $supported_types)) {
      return;
    }

    throw new InvalidConfigurationException(sprintf(
      'Requested to cast the value to an unsupported type "%s"',
      $this->configuration['cast_to']
    ));
  }

  /**
   * Returns the transformed value.
   *
   * The value is casted to the requested type, if defined in the configuration.
   *
   * @param mixed $value
   *   The field value before transformation.
   *
   * @return mixed
   *   The transformed field value.
   */
  protected function transformImportedValue(
    $value,
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    if (!empty($this->configuration['cast_to'])) {
      return $this->castValue($value, $this->configuration['cast_to']);
    }

    return $value;
  }

  /**
   * Returns the value casted to the given type.
   *
   * @param mixed $value
   *   The field value before transformation.
   * @param string $type
   *   The type to cast the value to.
   *
   * @return mixed
   *   The casted value.
   */
  protected function castValue($value, $type) {
    // No need to throw an exception if the type is unsupported; we validate the
    // type in the configuration validation method. Also, PHP will throw an
    // error as well.
    // Keep separate cases for types that are the same e.g. `int/integer`
    // etc. That behavior should be decided by PHP.
    switch ($type) {
      case 'int':
        return (int) $value;

      case 'integer':
        return (integer) $value;

      case 'bool':
        return (bool) $value;

      case 'boolean':
        return (boolean) $value;

      case 'float':
        return (float) $value;

      case 'double':
        return (double) $value;

      case 'real':
        return (real) $value;

      case 'string':
        return (string) $value;

      case 'array':
        return (array) $value;

      case 'object':
        return (object) $value;

      case 'unset':
        return (unset) $value;
    }
  }

}
