<?php

namespace Drupal\entity_sync\Plugin\EntitySync\FieldTransformer;

use Drupal\entity_sync\FieldTransformer\PluginBase;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\Validator\ValidatorInterface;

/**
 * Base class for transformers that select an entity ID based on a property.
 *
 * Supported configuration properties:
 * - entity_type_id: (string, required) The entity type ID of the entity that
 *   will be selected.
 * - property_name: (string, required) The name of the property or field that
 *   will be used to look up the entity.
 *
 * Implementations are required to implement the `propertyValue` method which
 * provides the property or field value that will be used to look up the
 * entity. Example implementations could be:
 * - The property must have a value equal to a fixed value defined in
 *   configuration.
 * - The property must have a value equal to that of another property/field of
 *   the local entity.
 * - The property must have a value equal to that of another property of the
 *   remote local entity.
 */
abstract class LocalEntityIdsByPropertyBase extends PluginBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Returns the value of the property that will be used to look up the entity.
   *
   * @param object $remote_entity
   *   The remote entity.
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param array $field_info
   *   The field info.
   *   See \Drupal\entity_sync\Import\Event\FieldMapping::fieldMapping.
   * @param array $context
   *   An associative array containing context related to the circumstances of
   *   the operation.
   *
   * @return mixed
   *   The property value.
   */
  abstract protected function propertyValue(
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  );

  /**
   * Constructs a EntityIdPropertyBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Symfony\Component\Validator\Validator\ValidatorInterface $constraint_validator
   *   The constraint validator.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $constraint_manager
   *   The constraint plugin manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    ValidatorInterface $constraint_validator,
    PluginManagerInterface $constraint_manager,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $constraint_validator,
      $constraint_manager
    );

    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    ValidatorInterface $constraint_validator,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $constraint_validator,
      $container->get('plugin.manager.entity_sync_validation_constraint'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function validateConfiguration() {
    $this->validateConfigurationRequiredProperties([
      'entity_type_id',
      'property_name',
    ]);
  }

  /**
   * {@inheritdoc}
   */
  protected function transformImportedValue(
    $value,
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    $property_value = $this->propertyValue(
      $remote_entity,
      $local_entity,
      $field_info,
      $context
    );

    // To make this base class generic, we simply return the array containing
    // the IDs. Implementations can further transform the value to the format
    // expected by the field they target.
    return $this->entityTypeManager
      ->getStorage($this->configuration['entity_type_id'])
      ->getQuery()
      ->accessCheck(FALSE)
      ->condition($this->configuration['property_name'], $property_value)
      ->execute();
  }

}
