<?php

namespace Drupal\entity_sync\Plugin\EntitySync\FieldTransformer;

use Drupal\Core\Entity\ContentEntityInterface;

/**
 * Transformer that selects an entity ID based on a remote property value.
 *
 * Supported configuration properties:
 * - entity_type_id: (string, required) The entity type ID of the entity that
 *   will be selected.
 * - property_name: (string, required) The name of the property or field that
 *   will be used to look up the entity.
 * - remote_entity_property_name: (string, required) The name of the remote
 *   property the value of which will be used to look up the entity.
 *
 * @EntitySyncFieldTransformer(
 *   id = "local_entity_ids_by_remote_property"
 * )
 */
class LocalEntityIdsByRemoteProperty extends LocalEntityIdsByPropertyBase {

  /**
   * {@inheritdoc}
   */
  public function validateConfiguration() {
    parent::validateConfiguration();

    $this->validateConfigurationRequiredProperties([
      'remote_entity_property_name',
    ]);
  }

  /**
   * {@inheritdoc}
   *
   * @throws \RuntimeException
   *   If the configured property does not exist in the remote entity.
   */
  protected function propertyValue(
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    // Getting the value of a property that does not exist would return a NULL
    // value, and that could have unintended consequences i.e. loading entities
    // that have a NULL value for the property. We would want that to be
    // intentional i.e. the property to exist and to have a NULL value.
    $property_exists = property_exists(
      $remote_entity,
      $this->configuration['remote_entity_property_name']
    );
    if (!$property_exists) {
      throw new \RuntimeException(sprintf(
        'The "%s" property does not exist in the remote entity.',
        $this->configuration['remote_entity_property_name']
      ));
    }

    return $remote_entity->{$this->configuration['remote_entity_property_name']};
  }

  /**
   * {@inheritdoc}
   *
   * This plugin is intended for entity reference fields. We therefore convert
   * the value as required for such fields.
   */
  protected function transformImportedValue(
    $value,
    \stdClass $remote_entity,
    ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    $transformed_value = parent::transformImportedValue(
      $value,
      $remote_entity,
      $local_entity,
      $field_info,
      $context
    );
    if (!$transformed_value) {
      return NULL;
    }

    return array_map(
      function ($id) {
        return ['target_id' => $id];
      },
      $transformed_value
    );
  }

}
